##############################################################################
#
# Copyright (c) 2002 Nexedi SARL and Contributors. All Rights Reserved.
#                  Philippe Beaumont <pb@nexedi.com>
#                  Jean-Paul Smets <jp@smets.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
#
# This program as such is not intended to be used by end users. End
# users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

"""SQLIndex

"""

__version__ = '$Revision: 1.33 $'[11:-2]


import  re, string
import operator,warnings
from Globals import Persistent,DTMLFile
from zLOG import LOG, ERROR
from OFS.SimpleItem import SimpleItem
from Acquisition import Implicit, Explicit
from Products.PluginIndexes.common.ResultList import ResultList
from Products.PluginIndexes import PluggableIndex
from Products.PluginIndexes.common.util import parseIndexRequest

# May need some work for compatibility by BTrees
from BTrees.IIBTree import IITreeSet, IISet, union, intersection
from BTrees.convert import convert


#
#  here I need to import what is needed for a connexion
#
import Shared.DC.ZRDB.Connection
import Shared.DC.ZRDB.DA

import Products.ZSQLMethods.SQL

# What is THIS !!!!!!!!!!!!!!!!!!!!!!
# Since the writers of Zope are doing this, me too, but really very dangerous!!!
from types import *

# Don't know what it is for
# it seems that this never changes
_marker = []

class SQLIndex(PluggableIndex.PluggableIndex, Persistent,
     Implicit, SimpleItem):
    """SQL index.
    doit permettre de creer un index qui pourra transmettre une requete SQL a une BD.

    TODO
        - please implement better exception handling if you use exceptions
        - implement Catalog Interface
    """

    __implements__ = (PluggableIndex.PluggableIndexInterface,)

    meta_type='SQLIndex'

    manage_options= (
        {'label': 'Settings',
         'action': 'manage_main',
# for now no help, but this is the next thing to do
#         'help': ('SQLIndex','SQLIndex_Settings.stx')
         },
    )

    query_options = ["query","range"]
    operators = ['or','and']
    useOperator = 'and'

    def __init__(self, id, ignore_ex=None, call_methods=None, caller=None, extra=None):
        """Create an index

        The arguments are:

          'id' -- the name of the item attribute to index.  This is
          either an attribute name or a record key.

          'extra'  -- Record to keep additional parameters

        """

        self.id = id
        self.catalog = caller
        if extra:
           self.connection_id = extra.connection
           self.query= extra.query
           self.query_nb_row = extra.query_nb_row
           self.query_del = extra.query_del
           self.query_del_all = extra.query_del_all
           self.query_read = extra.query_read
           self.query_read_one = extra.query_read_one
        else:
           raise("Panic: no connexion, no queries")

        # experimental code for specifing the operator
        # No idea what for yet
        self.operators = ['or','and']
        self.useOperator = 'and'


    # Main Methods To Implement

    def index_object(self, documentId, obj, threshold=None):
        """ index an object 'obj' with integer id 'documentId'"""
        # Has the indexing succeeded?
        returnStatus = 0

        # this one is the SQL Method that can add row to the DB
        method = self.catalog.unrestrictedTraverse(self.query)

        # There is a important issue here: checking if the element is already catalogued
        # thus the GetEntryForObject function must be useful
        if self.getEntryForObject(documentId) != "":
            # object already in the DB
            # thus for now we can update the object
            self.unindex_object(documentId)


        # list of the arguments of the SQL Method
        # somehow I need to check if I have exactly all the corresponding values and react if not
        # for now, it put None if no value is given to an argument
        # This question is nonsense!!!
        arguments = method.arguments_src

        # now building the set of values
        kw = {}
        for arg in string.split(arguments,' '):
            if arg is 'documentId':
                kw[arg] = documentId
            else:
                kw[arg] = getattr(obj, arg, None)
        kw['documentId']=documentId

        # cataloguing the object in the DB
        # Should I test if the cataloguing succeeds or act as I act in the unindex_object function?
        # I think that what I have done is correct, since I have change to be able to update
        method(**kw)

        returnStatus = 1

        return returnStatus

    def unindex_object(self, documentId):
        """ Unindex the object with integer id 'documentId' and don't
        raise an exception if we fail """
        try:
            self.catalog.unrestrictedTraverse(self.query_del)(documentId)
        except:
            LOG('UnIndex', ERROR, 'Attempt to unindex nonexistent document'
                ' with id %s' % documentId)

    def numObjects(self):
        """ return number of indexed objects
        for a SQL DB, this should be something like select count(*) from...
        """
        return self.catalog.unrestrictedTraverse(self.query_nb_row)()[0][0]

    def referencedObjects(self):
        """Generate a list of IDs for which we have referenced objects."""
        datas = self.catalog.unrestrictedTraverse(self.query_read)()._data
        return datas

    def getEntryForObject(self, documentId, default=_marker):
        """Takes a document ID and returns all the information we have
        on that specific object."""
        data = self.catalog.unrestrictedTraverse(self.query_read_one)(documentId=documentId)._data
        if len(data) > 0:
            return data[0]
        else:
            return ""

    def clear(self):
        """ the goal of this function is to clear the SQL table in the DB."""
        self.catalog.unrestrictedTraverse(self.query_del_all)()

    def _apply_index(self, request, cid='', type=type, None=None):
        """
        I will give more accurate details later
        Apply the index to query parameters given in the request arg.

        The request argument should be a mapping object.

        If the request does not have a key which matches the "id" of
        the index instance, then None is returned.

        If the request *does* have a key which matches the "id" of
        the index instance, one of a few things can happen:

          - if the value is a blank string, None is returned (in
            order to support requests from web forms where
            you can't tell a blank string from empty).

          - if the value is a nonblank string, turn the value into
            a single-element sequence, and proceed.

          - if the value is a sequence, return a union search.

        If the request contains a parameter with the name of the
        column + '_usage', it is sniffed for information on how to
        handle applying the index.

        If the request contains a parameter with the name of the
        column = '_operator' this overrides the default method
        ('or') to combine search results. Valid values are "or"
        and "and".

        If None is not returned as a result of the abovementioned
        constraints, two objects are returned.  The first object is a
        ResultSet containing the record numbers of the matching
        records.  The second object is a tuple containing the names of
        all data fields used.

        """

        record = parseIndexRequest(request, self.id, self.query_options)
        if record.keys==None: return None

        r     = None
        opr   = None

        sql_method = self.catalog.unrestrictedTraverse(self.query_read)

        # experimental code for specifing the operator
        operator = record.get('operator',self.useOperator)
        if not operator in self.operators :
           raise RuntimeError,"operator not valid: %s" % operator

        # depending on the operator we use intersection or union
        if operator=="or":  set_func = union
        else:               set_func = intersection

        # Range parameter
        range_parm = record.get('range',None)
        if range_parm:
            # This part is not implemented yet
            pass

        if record.get('usage',None):
            # see if any usage params are sent to field
            opr = record.usage.lower().split(':')
            opr, opr_args=opr[0], opr[1:]

        if opr=="range":   # range search
            # This part is not implemented yet
            pass
        else: # not a range search
            for key in record.keys:
                sql_result=sql_method(**key)
                # This is where a lot of optimization will be needed in the future
                set=IISet(map(lambda x: int(tuple(x)[0]),sql_result.tuples()))
                if set is not None:
                    if type(set) is IntType:
                        set = IISet((set,))
                    r = set_func(r, set)

        if type(r) is IntType:  r=IISet((r,))
        if r is None:
            return IISet(), (self.id,)
        else:
            return r, (self.id,)


    # Implementation Details

    def SQLConnectionIDs(self):
        """Find SQL database connections in the current folder and above
        This function return a list of ids.
        """
        ids={}
        have_id=ids.has_key
        StringType=type('')

        while self is not None:
            if hasattr(self, 'objectValues'):
                for o in self.objectValues():
                    if (hasattr(o,'_isAnSQLConnection') and o._isAnSQLConnection
                        and hasattr(o,'id')):
                        id=o.id
                        if type(id) is not StringType: id=id()
                        if not have_id(id):
                            if hasattr(o,'title_and_id'): o=o.title_and_id()
                            else: o=id
                            ids[id]=id
            if hasattr(self, 'aq_parent'): self=self.aq_parent
            else: self=None

        ids=map(lambda item: (item[1], item[0]), ids.items())
        ids.sort()
        return ids


    def SQLMethodIDs(self):
        """Find Z SQL methods in the current folder and above
        This function return a list of ids.
        """
        ids={}
        have_id=ids.has_key
        StringType=type('')

        while self is not None:
            if hasattr(self, 'objectValues'):
                for o in self.objectValues("Z SQL Method"):
                    if hasattr(o,'id'):
                        id=o.id
                        if type(id) is not StringType: id=id()
                        if not have_id(id):
                            if hasattr(o,'title_and_id'): o=o.title_and_id()
                            else: o=id
                            ids[id]=id
            if hasattr(self, 'aq_parent'): self=self.aq_parent
            else: self=None

        ids=map(lambda item: (item[1], item[0]), ids.items())
        ids.sort()
        return ids



    def manage_setPreferences(self,connection, query, extra,
                               REQUEST=None,RESPONSE=None,URL2=None):
        """ preferences of SQLIndex """

        # changing the connection
        if self.connection_id != connection:
            self.connection_id    = connection
        # changing the query that allows insertion into DB
        if self.query != query:
            self.query    = query
        # changing the query that allows deletion into DB
        if self.query_del != extra.query_del:
            self.query_del   = extra.query_del
        # changing the query that allows deletion of all tuples in DB
        if self.query_del_all != extra.query_del_all:
            self.query_del_all   = extra.query_del_all
        #changing the query that allows counting of the cataloged objects
        if self.query_nb_row != extra.query_nb_row:
            self.query_nb_row    = extra.query_nb_row
        #changing the query that allows reading all the cataloged objects
        if self.query_read != extra.query_read:
            self.query_read    = extra.query_read
        #changing the query that allows reading one of the cataloged objects
        if self.query_read_one != extra.query_read_one:
            self.query_read_one    = extra.query_read_one

        if RESPONSE:
            RESPONSE.redirect(URL2 + '/manage_main?manage_tabs_message=Preferences%20saved')

    def test_conn(self):
        """
            Test conn
        """
        return self.getattr('toto',None)

    manage_workspace  = DTMLFile("dtml/manageSQLIndex",globals())

manage_addSQLIndexForm = DTMLFile('dtml/addSQLIndex', globals())

def manage_addSQLIndex(self, id, extra=None, REQUEST=None, RESPONSE=None, URL3=None):
    """Add a SQL index"""
    return self.manage_addIndex(id, 'SQLIndex', extra, REQUEST, RESPONSE, URL3)
